@@ -1,7 +1,8 @@
-use crate::{ForcepError, MetaDb, Metadata, Result};
+use crate::{mem_cache::MemCache, ForcepError, MetaDb, Metadata, Result};
 use bytes::Bytes;
 use std::io;
 use std::path;
+use std::result;
 use tokio::fs as afs;
 
 /// Creates a writeable and persistent temporary file in the path provided, returning the path and
@@ -22,19 +23,32 @@ async fn tempfile(dir: &path::Path) -> Result<(afs::File, path::PathBuf)> {
 pub(crate) struct Options {
     pub(crate) path: path::PathBuf,
     pub(crate) dir_depth: u8,
-    // TODO: implement below option
-    pub(crate) save_last_access: bool,
+    pub(crate) track_access: bool,
+
+    // maximum size of the in-memory lru in bytes
+    pub(crate) lru_size: usize,
 
     // read and write buffer sizes
     pub(crate) rbuff_sz: usize,
     pub(crate) wbuff_sz: usize,
 }
 
-/// The main component of `forceps`, acts as the API for interacting with the on-disk API.
+/// The main component of `forceps`, and  acts as the API for interacting with the on-disk cache.
 ///
-/// This structure exposes `read`, `write`, and misc metadata operations. `read` and `write` are
-/// both async, whereas all metadata operations are sync. See [`CacheBuilder`](crate::CacheBuilder)
-/// for all customization options.
+/// This structure includes the async `read`, `write`, and `remove` operations which are the basic
+/// operations of the cache. It also includes some misc functions to interact with metadata and
+/// evict items from the cache.
+///
+/// # Eviction
+///
+/// This cache can evict items with a number of different eviction algorithms. To see more, see
+/// [`evict_with`] and the [`evictors`] module.
+///
+/// # Memory Cache
+///
+/// An in-memory cache can be optionally enabled as a layer over the regular on-disk cache. The
+/// memcache provides fast `HIT`s for recently used entries, circumventing filesystem operations
+/// altogether. To enable, use the [`CacheBuilder`]`::memory_lru_max_size` method.
 ///
 /// # Examples
 ///
@@ -49,9 +63,14 @@ pub(crate) struct Options {
 ///     .unwrap();
 /// # }
 /// ```
+///
+/// [`evict_with`]: #method.evict_with
+/// [`evictors`]: crate::evictors
+/// [`CacheBuilder`]: crate::CacheBuilder
 #[derive(Debug)]
 pub struct Cache {
     meta: MetaDb,
+    mem: MemCache,
     opts: Options,
 }
 
@@ -89,6 +108,7 @@ impl Cache {
         meta_path.push("index");
         Ok(Self {
             meta: MetaDb::new(&meta_path)?,
+            mem: MemCache::new(opts.lru_size),
             opts,
         })
     }
@@ -112,6 +132,15 @@ impl Cache {
         buf
     }
 
+    /// Tracks the access for a cache entry if the option is enabled
+    #[inline]
+    fn track_access_for(&self, k: &[u8]) -> Result<()> {
+        if self.opts.track_access {
+            self.meta.track_access_for(k)?;
+        }
+        Ok(())
+    }
+
     /// Reads an entry from the database, returning a vector of bytes that represent the entry.
     ///
     /// # Not Found
@@ -119,6 +148,12 @@ impl Cache {
     /// If the entry is not found, then it will return
     /// `Err(`[`Error::NotFound`](ForcepError::NotFound)`)`.
     ///
+    /// # Metadata
+    ///
+    /// This function will *not* perform a metadata read or write **unless** the `track_access`
+    /// build option is set. If the option is set, then it will perform a blocking read/write to
+    /// write new values to track the last access time and the total hits.
+    ///
     /// # Examples
     ///
     /// ```rust
@@ -138,9 +173,15 @@ impl Cache {
     /// ```
     pub async fn read<K: AsRef<[u8]>>(&self, key: K) -> Result<Bytes> {
         use tokio::io::AsyncReadExt;
+        let k = key.as_ref();
+
+        // look in the memory cache to see if it's there and return if it is
+        if let Some(val) = self.mem.get(k) {
+            return self.track_access_for(k).map(|_| val);
+        }
 
         let file = {
-            let path = self.path_from_key(key.as_ref());
+            let path = self.path_from_key(k);
             afs::OpenOptions::new()
                 .read(true)
                 .open(&path)
@@ -160,7 +201,11 @@ impl Cache {
             .read_to_end(&mut buf)
             .await
             .map_err(ForcepError::Io)?;
-        Ok(Bytes::from(buf))
+
+        self.track_access_for(k)?;
+        let bytes = Bytes::from(buf);
+        self.mem.put(k, Bytes::clone(&bytes));
+        Ok(bytes)
     }
 
     /// Writes an entry with the specified key to the cache database. This will replace the
@@ -207,6 +252,9 @@ impl Cache {
             .await
             .map_err(ForcepError::Io)?;
 
+        if !self.mem.is_nil() {
+            self.mem.put(key, Bytes::from(Vec::from(value)));
+        }
         self.meta.insert_metadata_for(key, value)
     }
 
@@ -326,6 +374,21 @@ impl Cache {
     pub fn metadata_iter(&self) -> impl Iterator<Item = Result<(Vec<u8>, Metadata)>> {
         self.meta.metadata_iter()
     }
+
+    /// Runs the specified eviction algorithm over this instance cache instance.
+    ///
+    /// Eviction algorithms will remove items out of the cache until certain a condition has been
+    /// met, usually a size requirement. See the [`evictors`] module for more information and
+    /// examples.
+    ///
+    /// [`evictors`]: crate::evictors
+    #[inline]
+    pub async fn evict_with<E>(&self, evictor: E) -> result::Result<(), E::Err>
+    where
+        E: crate::evictors::Evictor,
+    {
+        evictor.evict(self).await
+    }
 }
 
 #[cfg(test)]
@@ -355,6 +418,21 @@ mod test {
         cache.remove(&b"CACHE_KEY").await.unwrap();
     }
 
+    #[tokio::test]
+    async fn tracking_test() {
+        let cache = CacheBuilder::default()
+            .track_access(true)
+            .build()
+            .await
+            .unwrap();
+
+        cache.write(b"CACHE_KEY", b"Hello World").await.unwrap();
+        for _ in 0..100 {
+            cache.read(b"CACHE_KEY").await.unwrap();
+        }
+        assert_eq!(cache.read_metadata(b"CACHE_KEY").unwrap().get_hits(), 100);
+    }
+
     #[tokio::test]
     async fn read_metadata() {
         let cache = default_cache().await;
